
#include <iostream>
#include <memory>

struct Base
{
    virtual void print() const noexcept
    {
        std::cout << "Base (const)" << std::endl;
    }

    virtual void print() noexcept
    {
        std::cout << "Base" << std::endl;
    }

    virtual ~Base() = default;
};


struct Derived : Base
{
    virtual void print() const noexcept override
    {
        std::cout << "Derived (const)" << std::endl;
    }

    virtual void print() noexcept override
    {
        std::cout << "Derived" << std::endl;
    }
};


struct Fake
{
    virtual void print() const noexcept;
    //virtual void print() noexcept;
};


int main()
{
    {
        auto base{ std::make_unique<Base>() };
        base->print();

        auto derived{ std::make_unique<Derived>() };
        derived->print();
    }
    {
        auto base{ std::make_shared<Base>() };
        base->print();

        auto derived{ std::make_shared<Derived>() };
        derived->print();

        auto basePtr2Derived{ std::static_pointer_cast<Base>(derived) };
        basePtr2Derived->print();

        if (auto downcasted{ std::dynamic_pointer_cast<Derived>(basePtr2Derived) }; downcasted)
        {
            downcasted->print();
        }

        auto constBase{ std::const_pointer_cast<const Base>(base) };
        constBase->print();

        auto reinterpretedDerived{ std::reinterpret_pointer_cast<Fake>(derived) };
        reinterpretedDerived->print();
    }

    return 0;
}
